/***********************************************************************************/
/**	\file				CRender.cpp
**	\brief				Implementation zur Klasse CRender
*************************************************************************************
**	Autor:				Christian Roesch
**
**	Erstelldatum:		??.??.2002
**	letzte Aenderung:	??.??.2002
*************************************************************************************
**	-tut nichts-
**
************************************************************************************/


// includes
#include "CRender.h"

#include <fstream>

#include "CIObjekt.h"
#include "DRGBColor.h"

// Methoden-Definitionen
CRender::CRender( CRenderTarget& rt )
:
m_rt( rt )
{
    int siCount;
    float fAngle= 0.0f;
    float fAngleAdd= 3.1415f / ( c_siNrOfSectors * ( c_siNrOfRands + 1 ) );
    for ( siCount= 0;
          siCount < ( c_siNrOfSectors * ( c_siNrOfRands + 1 ) );
          siCount++, fAngle+= fAngleAdd )
    {
        m_fDirX[ siCount ]= sin (fAngle);
        m_fDirY[ siCount ]= cos (fAngle);
    }

    m_fPixelPerTile= 32;
    m_siTileShift= 5;

    if ( m_rt.GetWidth() <= 256)
    {
        m_fPixelPerTile= 16;
        m_siTileShift= 4;
    }
    if ( m_rt.GetWidth() <= 128)
    {
        m_fPixelPerTile= 8;
        m_siTileShift= 3;
    }
    m_siNrOfTileX= m_rt.GetWidth() >> m_siTileShift;
    m_siNrOfTileY= m_rt.GetHeight() >> m_siTileShift;

    m_pol= new CObjektList[ m_siNrOfTileX * m_siNrOfTileY * c_siNrOfSectors ];
}

CRender::~CRender()
{
    delete[] m_pol;
}

void CRender::Render( CObjektList& ol,
                      DVektor& mid,
                      DVektor& mat1,
                      DVektor& mat2,
                      float fSize )
{
    //std::ofstream ofs( "log.txt");

    int siScreenX;
    int siScreenY;
    int siOffset;
    DRay ray;
    DVektor& act= ray.m_Start;
    DWCOLOR* pdwc= m_rt.GetTargetPointer();

    float fDistPos;
    float fDistNeg;
    int siObjPos;
    int siObjNeg;

    mat1.m_fX*= fSize / float( m_rt.GetHeight() );
    mat1.m_fY*= fSize / float( m_rt.GetHeight() );
    mat2.m_fX*= fSize / float( m_rt.GetHeight() );
    mat2.m_fY*= fSize / float( m_rt.GetHeight() );

    mat1.m_fX*= m_rt.GetRatio();
    mat1.m_fY*= m_rt.GetRatio();

    mid.m_fX-= 0.5f * (float)( m_rt.GetWidth() - 1 )* mat1.m_fX +
               0.5f * (float)( m_rt.GetHeight() - 1 )* mat2.m_fX;
    mid.m_fY-= 0.5f * (float)( m_rt.GetWidth() - 1 )* mat1.m_fY +
               0.5f * (float)( m_rt.GetHeight() - 1 )* mat2.m_fY ;
/******************************************************************************/
// precreate Objekt lists
    DVektor pos;
    DVektor linestart;

    linestart.m_fX= mid.m_fX + 0.5f *
                               ( m_fPixelPerTile - 1 ) *
                               ( mat1.m_fX + mat2.m_fX );
    linestart.m_fY= mid.m_fY + 0.5f *
                               ( m_fPixelPerTile - 1 ) *
                               ( mat1.m_fY + mat2.m_fY );

    float fRad;

    fRad= ( linestart.m_fX - mid.m_fX ) * ( linestart.m_fX - mid.m_fX );
    fRad+= ( linestart.m_fY - mid.m_fX ) * ( linestart.m_fY - mid.m_fY );

    fRad= sqrt( fRad );

    siOffset= 0;
    for( siScreenY= 0;
         siScreenY < m_siNrOfTileY;
         siScreenY++, linestart.m_fX+= mat2.m_fX * m_fPixelPerTile, linestart.m_fY+= mat2.m_fY * m_fPixelPerTile )
    {
        for( siScreenX= 0, pos.m_fX= linestart.m_fX, pos.m_fY= linestart.m_fY;
             siScreenX < m_siNrOfTileX;
             siScreenX++, pos.m_fX+= mat1.m_fX * m_fPixelPerTile, pos.m_fY+= mat1.m_fY * m_fPixelPerTile )
        {
            int siAngle;
            for ( siAngle= 0; siAngle < c_siNrOfSectors; siAngle++)
            {
                m_pol[siOffset].Clear();
                DVektor norm1;
                DVektor norm2;
                norm1.m_fY= m_fDirX[ siAngle * ( c_siNrOfRands + 1 ) ];
                norm1.m_fX= -m_fDirY[ siAngle * ( c_siNrOfRands + 1 ) ];
                norm2.m_fY= m_fDirX[ siAngle * ( c_siNrOfRands + 1 ) + c_siNrOfRands];
                norm2.m_fX= -m_fDirY[ siAngle * ( c_siNrOfRands + 1 ) + c_siNrOfRands];
/******************************************************************************/
// iterate over any objects
    int siObjekt;
    for( siObjekt= 0;
         siObjekt < ol.GetNrOf();
         siObjekt++ )
    {
        CIObjekt* pobj= ol.GetObjekt( siObjekt );
        if ( pobj->Visible( pos,
                            fRad,
                            norm1,
                            norm2 ) )
        {
            m_pol[siOffset].AddObjekt( pobj );
        }
    }
/******************************************************************************/
                siOffset++;
            }
        }
    }

/******************************************************************************/
// fill Target
    siOffset= 0;
    for( siScreenY= 0;
         siScreenY < m_rt.GetHeight();
         siScreenY++, mid.m_fX+= mat2.m_fX, mid.m_fY+= mat2.m_fY )
    {
        for( siScreenX= 0, act.m_fX= mid.m_fX, act.m_fY= mid.m_fY;
             siScreenX < m_rt.GetWidth();
             siScreenX++, act.m_fX+= mat1.m_fX, act.m_fY+= mat1.m_fY )
        {
/******************************************************************************/
            DRGBColor rgbTmp;
            DRGBColor rgb;

            memset( (void*)&rgb, 0, sizeof(rgb));

            int siList= siScreenY >> m_siTileShift;
            siList*= m_siNrOfTileX;
            siList+= siScreenX >> m_siTileShift;
            siList*= c_siNrOfSectors;

            // iterate over all angles
            int siAngle;
            for ( siAngle= 0; siAngle < c_siNrOfSectors; siAngle++, siList++ )
            {
                if ( m_pol[siList].GetNrOf() < 2)
                {
                    if ( m_pol[siList].GetNrOf() == 0)
                    {
                        continue;
                    }

                    {
                        int siRand= rand() & c_siNrOfRands;
                        ray.m_Dir.m_fX= m_fDirX[ siAngle * ( c_siNrOfRands + 1 ) + siRand ];
                        ray.m_Dir.m_fY= m_fDirY[ siAngle * ( c_siNrOfRands + 1 ) + siRand ];
                    }

                    float fDist;
                    if ( m_pol[siList].GetObjekt( 0 )->Intersect( &ray, fDist ) )
                    {
                        m_pol[siList].GetObjekt( 0 )->GetColor( fabs(fDist), rgbTmp );

                        rgb.m_iRed+= rgbTmp.m_iRed;
                        rgb.m_iGreen+= rgbTmp.m_iGreen;
                        rgb.m_iBlue+= rgbTmp.m_iBlue;
                    }
                    continue;
                }

                {
                    int siRand= rand() & c_siNrOfRands;
                    ray.m_Dir.m_fX= m_fDirX[ siAngle * ( c_siNrOfRands + 1 ) + siRand ];
                    ray.m_Dir.m_fY= m_fDirY[ siAngle * ( c_siNrOfRands + 1 ) + siRand ];
                }

/******************************************************************************/
// iterate over any objects
    fDistPos= 1000000.0f;
    fDistNeg= 1000000.0f;
    siObjPos= -1;
    siObjNeg= -1;

    float fDist;
    int siObjekt;
    for( siObjekt= 0;
         siObjekt < m_pol[siList].GetNrOf();
         siObjekt++ )
    {
        if (! m_pol[siList].GetObjekt( siObjekt )->Intersect( &ray, fDist ) )
        {
            continue;
        }
        if( fDist > 0.0f )
        {
            if( fDist < fDistPos )
            {
                fDistPos= fDist;
                siObjPos= siObjekt;
            }
            continue;
        }

        if( fDist < 0.0f )
        {
            fDist= -fDist;
            if( fDist < fDistNeg )
            {
                fDistNeg= fDist;
                siObjNeg= siObjekt;
            }
            continue;
        }

        fDistPos= fDist;
        siObjPos= siObjekt;
        siObjNeg= -1;
        break;
    }
/******************************************************************************/
                if ( siObjPos != -1 )
                {
                    m_pol[siList].GetObjekt( siObjPos )->GetColor( fDistPos, rgbTmp );

                    rgb.m_iRed+= rgbTmp.m_iRed;
                    rgb.m_iGreen+= rgbTmp.m_iGreen;
                    rgb.m_iBlue+= rgbTmp.m_iBlue;
                }
                if ( siObjNeg != -1 )
                {
                    m_pol[siList].GetObjekt( siObjNeg )->GetColor( fDistNeg, rgbTmp );

                    rgb.m_iRed+= rgbTmp.m_iRed;
                    rgb.m_iGreen+= rgbTmp.m_iGreen;
                    rgb.m_iBlue+= rgbTmp.m_iBlue;
                }
            }
            pdwc[siOffset].ucRed= rgb.m_iRed >> c_siColorShift;
            pdwc[siOffset].ucGreen= rgb.m_iGreen >> c_siColorShift;
            pdwc[siOffset].ucBlue= rgb.m_iBlue >> c_siColorShift;
            rand();
            siOffset++;
/******************************************************************************/
        }
    }

}

/************************************************************************************
**	Ende der Datei:		CRender.cpp
************************************************************************************/

